/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Line
Version:   1.0
Status:    Stable
ID:        ID_LINE
Category:  GUI
Date:      January 2002
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2002.  All rights reserved.
Keywords:  line, horizontal, vertical, angle, draw, graphic
Short:     Draws lines to graphical areas.

###DESCRIPTION###
<p>The Line class is used to draw single and multiple lines to graphics
areas.  The class supports individual colour settings for each line,
scaling and variable line thickness.  Lines can be drawn in a sequence
of points and support is provided for offsets, fixed coordinates and
relative coordinates.  Here are four DML examples that demonstrate different
ways to use the Line class:</p>

<pre>
  &lt;line points="(100,100) (150,76) (187,560) (0,0)" col0="#ffffff"
    col1="#ff0000" col2="#00ff00"/&gt;

  &lt;line p0="1,1" p1="!100,!400" colour="#ffffff"/&gt;

  &lt;line p0="50%,50%" p1="!25%,!25%" colour="#ffffff"/&gt;

  &lt;line x0="=[logo.x]-10" y0="0" x1="=[logo.x]-10"
    y1="=[systemdisplay.height]-1" colour="#FFFFFF"/&gt;
</pre>

<p>Because the Line class produces graphical objects, they should always be
initialised to a drawable container, such as a Render object.  For more
detail on how to use the Line class, please refer to the documentation for
the individual fields.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/functions/strtoint.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_LINE 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_LINE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2002. All rights reserved.",
   "February 2002",
   "Line"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR LineClass = 0;

/****************************************************************************
** Line class.
*/

struct Point {
   FLOAT XCoord;
   FLOAT YCoord;
   WORD Flags;
   struct RGB Colour;
};

struct Line {
   OBJECT_HEADER
   DOUBLE   Scale;
   OBJECTID DrawableID;
   LONG     Visible;
   LONG     Frame;

   PRIVATE_FIELDS
   WORD TotalPoints;
   WORD Opacity;
   WORD   *Frames;
   WORD   TotalFrames;
   WORD   CurrentFrame;
   struct BitmapPen *Pen;
   struct RGB Colour;
   struct Point Points[20];
};

#define LF_RELATIVE_X 0x0001
#define LF_OFFSET_X   0x0002
#define LF_RELATIVE_Y 0x0004
#define LF_OFFSET_Y   0x0008
#define LF_COLOUR     0x0010
#define LF_BREAK      0x0020

#define LF_XFLAGS (LF_RELATIVE_X|LF_OFFSET_X)
#define LF_YFLAGS (LF_RELATIVE_Y|LF_OFFSET_Y)

static ERROR GET_ColourRGB(struct Line *, struct RGB *);
static ERROR GET_Opacity(struct Line *, FLOAT *);
static ERROR GET_Thickness(struct Line *, LONG *);

static ERROR SET_Colour(struct Line *, STRING);
static ERROR SET_ColourRGB(struct Line *, struct RGB *);
static ERROR SET_Frames(struct Line *, STRING);
static ERROR SET_Opacity(struct Line *, FLOAT *);
static ERROR SET_Points(struct Line *, STRING);
static ERROR SET_Scale(struct Line *, DOUBLE *);
static ERROR SET_Thickness(struct Line *, LONG *);
static ERROR SET_X(struct Line *, struct Variable *);
static ERROR SET_XEnd(struct Line *, struct Variable *);
static ERROR SET_Y(struct Line *, struct Variable *);
static ERROR SET_YEnd(struct Line *, struct Variable *);

static struct FieldArray LineFields[] = {
   { "Scale",     0, FDF_DOUBLE|FDF_RW,   0,         NULL, SET_Scale },
   { "Drawable",  0, FDF_OBJECTID|FDF_RI, ID_RENDER, NULL, NULL },
   { "Visible",   0, FDF_LONG|FDF_RW,     0,         NULL, NULL },
   { "Frame",     0, FDF_LONG|FDF_RW,     0,         NULL, NULL },
   /*** Virtual Fields ***/
   { "Colour",    0, FDF_STRING|FDF_W,  0, NULL,          SET_Colour },
   { "ColourRGB", 0, FDF_POINTER|FDF_R, 0, GET_ColourRGB, SET_ColourRGB },
   { "Frames",    0, FDF_STRING|FDF_W,  0, NULL,          SET_Frames },
   { "Opacity",   0, FDF_FLOAT|FDF_RW,  0, GET_Opacity,   SET_Opacity },
   { "Points",    0, FDF_STRING|FDF_W,  0, NULL,          SET_Points },
   { "Thickness", 0, FDF_LONG|FDF_RW,   0, GET_Thickness, SET_Thickness },
   { "X",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W, 0, NULL, SET_X },
   { "XCoord",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W, 0, NULL, SET_X },
   { "XEnd",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W, 0, NULL, SET_XEnd },
   { "Y",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W, 0, NULL, SET_Y },
   { "YCoord",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W, 0, NULL, SET_Y },
   { "YEnd",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W, 0, NULL, SET_YEnd },
   END_FIELD
};

static ERROR LINE_Free(struct Line *, APTR);
static ERROR LINE_Hide(struct Line *, APTR);
static ERROR LINE_Init(struct Line *, APTR);
static ERROR LINE_NewObject(struct Line *, APTR);
static ERROR LINE_SetUnlistedField(struct Line *, struct acSetUnlistedField *);
static ERROR LINE_Show(struct Line *, APTR);

static struct ActionArray LineActions[] = {
  { AC_Free,             LINE_Free },
  { AC_Hide,             LINE_Hide },
  { AC_Init,             LINE_Init },
  { AC_NewObject,        LINE_NewObject },
  { AC_SetUnlistedField, LINE_SetUnlistedField },
  { AC_Show,             LINE_Show },
  { NULL, NULL }
};

static WORD ReadByte(STRING);
static ERROR ReadColour(STRING, LONG *, LONG *, LONG *);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&LineClass, NULL,
      FID_BaseClassID|TLONG, ID_LINE,
      FID_Version|TFLOAT,    VER_LINE,
      FID_Name|TSTRING,      "Line",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      LineActions,
      FID_Fields|TPTR,       LineFields,
      FID_Flags|TLONG,       CLF_PRIVATEONLY,
      FID_Size|TLONG,        sizeof(struct Line),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (LineClass)  { Action(AC_Free, LineClass, NULL); LineClass  = 0; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawLine()
*/

static void DrawLine(struct Line *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   LONG i;

   if (Self->TotalPoints < 2) return;

   if (Self->Visible IS FALSE) return;

   /*** Frame testing ***/

   if ((Self->Frame) OR (Self->Frames)) {
      if (Self->Frame) {
         if (Drawable->Frame != Self->Frame) return;
      }
      else if (Self->Frames) {
         if ((Drawable->Frame > Self->TotalFrames) OR (Self->Frames[Drawable->Frame] < 1)) return;
      }
   }

   Self->Pen->Bitmap = Bitmap;

   for (i=0; i < Self->TotalPoints-1; i++) {
      /*** If the next line is set for a break, increase the counter by 1 ***/

      if (Self->Points[i+1].Flags & LF_BREAK) {
         i++;
         if (i >= Self->TotalPoints-1) break;
      }

      /*** X1 ***/

      if (Self->Points[i].Flags & LF_OFFSET_X) {
         if (Self->Points[i].Flags & LF_RELATIVE_X) Self->Pen->XCoord = Drawable->Width - (Self->Points[i].XCoord * Drawable->Width / 100);
         else Self->Pen->XCoord = Drawable->Width - Self->Points[i].XCoord - 1;
      }
      else if (Self->Points[i].Flags & LF_RELATIVE_X) Self->Pen->XCoord = Self->Points[i].XCoord * Drawable->Width / 100;
      else Self->Pen->XCoord = Self->Points[i].XCoord;

      /*** X2 ***/

      if (Self->Points[i+1].Flags & LF_OFFSET_X) {
         if (Self->Points[i+1].Flags & LF_RELATIVE_X) Self->Pen->XEnd = Drawable->Width - (Self->Points[i+1].XCoord * Drawable->Width / 100);
         else Self->Pen->XEnd = Drawable->Width - Self->Points[i+1].XCoord - 1;
      }
      else if (Self->Points[i+1].Flags & LF_RELATIVE_X) Self->Pen->XEnd = Self->Points[i+1].XCoord * Drawable->Width / 100;
      else Self->Pen->XEnd = Self->Points[i+1].XCoord;

      /*** Y1 ***/

      if (Self->Points[i].Flags & LF_OFFSET_Y) {
         if (Self->Points[i].Flags & LF_RELATIVE_Y) Self->Pen->YCoord = Drawable->Height - (Self->Points[i].YCoord * Drawable->Height / 100);
         else Self->Pen->YCoord = Drawable->Height - Self->Points[i].YCoord - 1;
      }
      else if (Self->Points[i].Flags & LF_RELATIVE_Y) Self->Pen->YCoord = Self->Points[i].YCoord * Drawable->Height / 100;
      else Self->Pen->YCoord = Self->Points[i].YCoord;

      /*** Y2 ***/

      if (Self->Points[i+1].Flags & LF_OFFSET_Y) {
         if (Self->Points[i+1].Flags & LF_RELATIVE_Y) Self->Pen->YEnd = Drawable->Height - (Self->Points[i+1].YCoord * Drawable->Height / 100);
         else Self->Pen->YEnd = Drawable->Height - Self->Points[i+1].YCoord - 1;
      }
      else if (Self->Points[i+1].Flags & LF_RELATIVE_Y) Self->Pen->YEnd = Self->Points[i+1].YCoord * Drawable->Height / 100;
      else Self->Pen->YEnd = Self->Points[i+1].YCoord;

      /*** Colour ***/

      if (Self->Points[i].Flags & LF_COLOUR) {
         Self->Pen->Red   = Self->Points[i].Colour.Red;
         Self->Pen->Green = Self->Points[i].Colour.Green;
         Self->Pen->Blue  = Self->Points[i].Colour.Blue;
      }
      else {
         Self->Pen->Red   = Self->Colour.Red;
         Self->Pen->Green = Self->Colour.Green;
         Self->Pen->Blue  = Self->Colour.Blue;
      }

      /*** Scale ***/

      if ((Self->Scale != 1) AND (Self->Scale != 0)) {
         Self->Pen->XCoord *= Self->Scale;
         Self->Pen->YCoord *= Self->Scale;
         Self->Pen->XEnd   *= Self->Scale;
         Self->Pen->YEnd   *= Self->Scale;
      }

      Action(AC_Draw, Self->Pen, NULL);
   }
}

/*****************************************************************************
** Line: Free
*/

static ERROR LINE_Free(struct Line *Self, APTR Void)
{
   if (Self->DrawableID) {
      OBJECTPTR drawable;
      if (AccessObject(Self->DrawableID, 2000, &drawable) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, drawable, Self, NULL);
         ReleaseObject(drawable);
      }
   }

   if (Self->Pen) {
      Action(AC_Free, Self->Pen, NULL);
      Self->Pen = NULL;
   }

   if (Self->Frames) {
      FreeMemory(Self->Frames);
      Self->Frames = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Hides the graphics of a line object.
###END###

*****************************************************************************/

static ERROR LINE_Hide(struct Line *Self, APTR Void)
{
   if (Self->Visible IS TRUE) {
      Self->Visible = FALSE;
      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Line: Init
*/

static ERROR LINE_Init(struct Line *Self, APTR Void)
{
   OBJECTPTR drawable;

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (OwnerID) Self->DrawableID = OwnerID;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, drawable, Self, &DrawLine);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_Init, ERR_ExclusiveDenied));
}

/*****************************************************************************
** Line: NewObject
*/

static ERROR LINE_NewObject(struct Line *Self, APTR Void)
{
   if (NewPrivateObject(ID_PEN, NF_CHILD, &Self->Pen) IS ERR_Okay) {
      SetFields(Self->Pen, FID_PenShape|TLONG, PEN_LINE, TAGEND);
      Self->Visible = TRUE;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_NewObject));
}

/*****************************************************************************
** Line: SetUnlistedField
*/

static ERROR LINE_SetUnlistedField(struct Line *Self, struct acSetUnlistedField *Args)
{
   STRING str;
   WORD offset;
   LONG i, j;

   if ((!Args) OR (!Args->Field) OR (!Args->Value)) {
      return(ObjectError(Self, ERH_SetUnlistedField, ERR_Args));
   }

   if ((Args->Field[0] IS 'p') OR (Args->Field[0] IS 'P')) {
      i = StrToInt(Args->Field+1);
      str = Args->Value;

      /*** X Coordinate ***/

      offset = FALSE;
      while ((*str) AND ((*str < '0') OR (*str > '9'))) {
         if (*str IS '!') offset = TRUE;
         str++;
      }
      if (!*str) return(ERR_BadData);

      Self->Points[i].Flags = NULL;
      if (offset IS TRUE) Self->Points[i].Flags |= LF_OFFSET_X;
      Self->Points[i].XCoord = StrToInt(str);

      while ((*str >= '0') AND (*str <= '9')) str++;
      if (*str IS NULL) return(ERR_BadData);
      else if (*str IS '%') Self->Points[i].Flags |= LF_RELATIVE_X;

      /*** Y Coordinate ***/

      offset = FALSE;
      while ((*str) AND ((*str < '0') OR (*str > '9'))) {
         if (*str IS '!') offset = TRUE;
         str++;
      }
      if (!*str) return(ERR_BadData);

      if (offset IS TRUE) Self->Points[i].Flags |= LF_OFFSET_Y;
      Self->Points[i].YCoord = StrToInt(str);

      while ((*str >= '0') AND (*str <= '9')) str++;
      if (*str IS NULL) return(ERR_BadData);
      else if (*str IS '%') Self->Points[i].Flags |= LF_RELATIVE_Y;
   }
   else if ((Args->Field[0] IS 'x') OR (Args->Field[0] IS 'X')) {
      /*** X Coordinate ***/
      i = StrToInt(Args->Field+1);
      Self->Points[i].Flags &= ~LF_XFLAGS;
      if (Args->Value[0] IS '!') {
         Self->Points[i].XCoord = StrToInt(Args->Value+1);
         Self->Points[i].Flags |= LF_OFFSET_X;
      }
      else Self->Points[i].XCoord = StrToInt(Args->Value);

      for (j=0; (Args->Value[j] != NULL) AND (Args->Value[j] != '%'); j++);
      if (Args->Value[j] IS '%') Self->Points[i].Flags |= LF_RELATIVE_X;
   }
   else if ((Args->Field[0] IS 'y') OR (Args->Field[0] IS 'Y')) {
      /*** Y Coordinate ***/
      i = StrToInt(Args->Field+1);
      Self->Points[i].Flags &= ~LF_YFLAGS;
      if (Args->Value[0] IS '!') {
         Self->Points[i].YCoord = StrToInt(Args->Value+1);
         Self->Points[i].Flags |= LF_OFFSET_Y;
      }
      else Self->Points[i].YCoord = StrToInt(Args->Value);

      for (j=0; (Args->Value[j] != NULL) AND (Args->Value[j] != '%'); j++);
      if (Args->Value[j] IS '%') Self->Points[i].Flags |= LF_RELATIVE_Y;
   }
   else if (StrCompare("col", Args->Field, 3, NULL) IS ERR_Okay) {
      for (i=0; (Args->Field[i]) AND ((Args->Field[i] < '0') OR (Args->Field[i] > '9')); i++);
      if (Args->Field[0]) {
         i = StrToInt(Args->Field+i);
         ReadColour(Args->Value, &Self->Points[i].Colour.Red, &Self->Points[i].Colour.Green, &Self->Points[i].Colour.Blue);
         Self->Points[i].Flags |= LF_COLOUR;
      }
   }
   else {
      DPrintF("!SetField()","Unrecognised field %s, value %s", Args->Field, Args->Value);
      return(ERR_Okay);
   }

   if (i+1 > Self->TotalPoints) Self->TotalPoints = i + 1;

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Displays the graphics of a line object.
###END###

*****************************************************************************/

static ERROR LINE_Show(struct Line *Self, APTR Void)
{
   if (Self->Visible IS FALSE) {
      Self->Visible = TRUE;
      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING str)
{
   WORD value;
   if (!*str) return(0);
   if ((*str >= '0') AND (*str <= '9')) value = (*str - '0') * 16;
   else if ((*str >= 'A') AND (*str <= 'F')) value = ((*str - 'A')+10) * 16;
   else if ((*str >= 'a') AND (*str <= 'f')) value = ((*str - 'a')+10) * 16;
   else value = 0;
   str++;
   if ((*str >= '0') AND (*str <= '9')) value += (*str - '0');
   else if ((*str >= 'A') AND (*str <= 'F')) value += (*str - 'A') + 10;
   else if ((*str >= 'a') AND (*str <= 'f')) value += (*str - 'a') + 10;
   return(value);
}

/*****************************************************************************
** Internal: ReadColour()
** Short:    Reads colours in hex and RGB formats.
*/

static ERROR ReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue)
{
   if (*Colour IS '#') {
      Colour++;
      *Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Blue  = ReadByte(Colour);
   }
   else {
      *Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Blue = StrToInt(Colour);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Defines the colour of the line in RGB format.
Type:   *RGB
Status: Read/Write

###DESCRIPTION###
<p>The colour of a line can be read or written in RGB format through this
field.</p>

<p>If you are creating multiple lines and need to set the colour of each
line individually, you need to use unlisted fields to do this.  Refer to the
<a href="#Points">Points</a> field for more detail.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct Line *Self, struct RGB *RGB)
{
   RGB->Red   = Self->Colour.Red;
   RGB->Green = Self->Colour.Green;
   RGB->Blue  = Self->Colour.Blue;
   return(ERR_Okay);
}

static ERROR SET_ColourRGB(struct Line *Self, struct RGB *RGB)
{
   Self->Colour.Red   = RGB->Red;
   Self->Colour.Green = RGB->Green;
   Self->Colour.Blue  = RGB->Blue;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for setting the colour of a Line.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>The colour of a Line can be declared by writing to this field.  The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of '#ff0000' or '255,0,0' would be valid.</p>

<p>If you are creating multiple lines and need to set the colour of each
line individually, you need to use unlisted fields to do this.  Refer to the
Points field for more detail.</p>

###SEE ALSO###
Field: Points

###END###

*****************************************************************************/

static ERROR SET_Colour(struct Line *Self, STRING Colour)
{
   if (!Colour) return(ERR_Failed);

   ReadColour(Colour, &Self->Colour.Red, &Self->Colour.Green, &Self->Colour.Blue);
   return(ERR_Okay);
}


/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Limits the drawing of a line to a single frame.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Setting the Frame field to any value other than zero will force a line to
be drawn only when the drawable's frame matches the specified value.  For
instance, if the drawable container has a frame setting of 2, and the line has
a frame of 1, then no graphics will be drawn to represent the line as the
numbers do not match.</p>

###SEE ALSO###
Field: Frames

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frames
Short:  Determines a sequence of frames that a line graphic may be drawn to.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>For long animation sequences you can set the Frames field to determine
individual frames that a line graphic should be drawn to.  For instance,
if a drawable has been assigned 20 animation frames and you want a line
graphic to appear in frames 5, 6, 7, 8, 17 and 19, you would use this
frame specification:</p>

<pre>
   5-8,17,19
</pre>

<p>Individual frame sets are separated by any non-numeric character such as
a comma, and ranges can be specified through the '-' character.</p>

<p>Setting the Frames field will override any setting that you may have
previously set in the <a href="#Frame">Frame</a> field.</p>

###SEE ALSO###
Field: Frame

###END###

*****************************************************************************/

static ERROR SET_Frames(struct Line *Self, STRING Value)
{
   STRING str;
   LONG frame, end;

   if ((!Value) OR (!(*Value))) return(ERR_Args);

   /*** Calculate the size of the frame array that we are going to need ***/

   Self->TotalFrames = 0;
   str = Value;
   while (*str) {
      if ((*str >= '0') AND (*str <= '9')) {
         frame = StrToInt(str);
         while ((*str >= '0') AND (*str <= '9')) str++;
         if (frame > Self->TotalFrames) Self->TotalFrames = frame;
      }
      else str++;
   }

   Self->TotalFrames++; /* Add an extra frame because we start from #1, not zero */

   /*** Allocate the frame array and record the sequence ***/

   if (AllocMemory((Self->TotalFrames+1) * sizeof(WORD), MEM_DATA, &Self->Frames, NULL) IS ERR_Okay) {
      str = Value;
      while (*str) {
         if ((*str >= '0') AND (*str <= '9')) {
            if ((frame = StrToInt(str)) > 0) {
               while ((*str >= '0') AND (*str <= '9')) str++;
               while ((*str) AND (*str <= 0x20)) str++;
               if (*str IS '-') {
                  while ((*str) AND ((*str < '0') OR (*str > '9'))) str++;
                  if ((end = StrToInt(str)) > 0) {
                     while (frame != end) {
                        Self->Frames[frame] = 1;
                        if (frame < end) frame++;
                        else frame--;
                     }
                     Self->Frames[frame] = 1;
                  }
                  while ((*str >= '0') AND (*str <= '9')) str++;
               }
               else Self->Frames[frame] = 1;
            }
            else while ((*str >= '0') AND (*str <= '9')) str++;
         }
         else str++;
      }

      Self->Frames[Self->TotalFrames] = -1; /* Sequence termination */
   }
   else return(ERR_AllocMemory);

   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Opacity
Short:  Determines the level of translucency applied to a line.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of a line graphic.  The
default setting is 100%, which means that the line will be solid.  Any
other value that you set here will alter the impact of a line graphic over
the destination drawable.  High values will retain the boldness of the line,
while low values can render it close to invisible.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw a line.</p>

###END###

****************************************************************************/

static ERROR GET_Opacity(struct Line *Self, FLOAT *Value)
{
   *Value = Self->Pen->Alpha * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_Opacity(struct Line *Self, FLOAT *Value)
{
   if (*Value < 0) SetFields(Self->Pen, FID_Alpha|TLONG, 0, TAGEND);
   else if (*Value > 100) SetFields(Self->Pen, FID_Alpha|TLONG, 255, TAGEND);
   else SetFields(Self->Pen, FID_Alpha|TLONG, (LONG)(*Value * 255 / 100), TAGEND);
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Points
Short:  This field provides an easy way to draw multiple and connected lines.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>This special field allows you to create multiple line definitions using
a single Line object.  This can help you save memory in cases where you might
otherwise need to create many Line objects to get the effect that you want.
Setting this field requires that you follow some basic rules and use the
correct symbols, or you may get unpredictable results.</p>

<p>Basic use requires that you specify the coordinates and dimensions for
each line in sets of four numbers - that is, the X and Y coordinates followed
by the XEnd and YEnd coordinates.  To aid you in your visualisation of these
numbers, you may insert brackets and commas to organise each line definition
into a group.  The following example illustrates the creation of a series of
line graphics:</p>

<pre>
  "(10,10) (50,50) (100,0) (85,150) (400,30) (34,300)"
</pre>

<p>This would result in five lines being drawn across the graphics area
from one point to the next.  If you need to set offsets or percentage
based values, then you will need to use the exclamation and/or percentage
symbols.  This next example shows all possible symbol combinations being
used:</p>

<pre>
  "(10,10) (50,50) (25%,33%) (!60,!25) (!103%,!95%)"
</pre>

<p>If you need to break the line so that it is not fully connected, you
can use the 'B' character to set a disconnection.  This next example shows
the previous set of lines being broken into two via use of the B
character:</p>

<pre>
  "(10,10) (50,50) B (25%,33%) (!60,!25) (!103%,!95%)"
</pre>

<p>If you need to set individual colours for each line, the unlisted field
mechanism needs to be used, and you need to know the correct index for
each line.  To demonstrate, setting the colour for the first line in
the list would require us to set the field 'col0'.  If you are using broken
lines then be aware that the index is based on the point settings, so in
our previous example the second line starts at point 3 and thus its colour
is accessed through 'col2'.  In situations where a specific colour has not
been set, the value in the <a href="#Colour">Colour</a> field is used as a
default.</p>

###END###

*****************************************************************************/

static ERROR SET_Points(struct Line *Self, STRING Points)
{
   WORD offset, i;

   i = 0;
   while (*Points) {
      Self->Points[i].Flags = NULL;

      /*** X Coordinate ***/

      offset = FALSE;
      while ((*Points) AND ((*Points < '0') OR (*Points > '9'))) {
         if (*Points IS '!') offset = TRUE;
         if ((*Points IS 'b') OR (*Points IS 'B')) Self->Points[i].Flags |= LF_BREAK;
         Points++;
      }
      if (!*Points) break;

      if (offset IS TRUE) Self->Points[i].Flags |= LF_OFFSET_X;
      Self->Points[i].XCoord = StrToInt(Points);
      while ((*Points >= '0') AND (*Points <= '9')) Points++;
      if (*Points IS NULL) break;
      else if (*Points IS '%') Self->Points[i].Flags |= LF_RELATIVE_X;

      /*** Y Coordinate ***/

      offset = FALSE;
      while ((*Points) AND ((*Points < '0') OR (*Points > '9'))) {
         if (*Points IS '!') offset = TRUE;
         Points++;
      }
      if (!*Points) break;

      if (offset IS TRUE) Self->Points[i].Flags |= LF_OFFSET_Y;
      Self->Points[i].YCoord = StrToInt(Points);
      while ((*Points >= '0') AND (*Points <= '9')) Points++;
      if (*Points IS '%') Self->Points[i].Flags |= LF_RELATIVE_Y;

      i++;
   }

   Self->TotalPoints = i;

   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Scale
Short:  Adjusts the scale of the Line.
Type:   DOUBLE
Status: Read/Set

###DESCRIPTION###
<p>The scale of a Line can be adjusted by writing a value to this line.  By
default the Scale is set to 1, which tells the object not to use the scaling
multiplier.  In all other cases, the coordinates of the line will be
multiplied by the Scale.  For example, a line from (25,60) to (90,104) with
a scale of 0.5 would be drawn as a line from (12,30) to (45,52).</p>

<p>Writing to the Scale field results in an immediate graphics update.
Negative values are not permitted.</p>

###END###

****************************************************************************/

static ERROR SET_Scale(struct Line *Self, DOUBLE *Scale)
{
   if ((*Scale >= 0) AND (*Scale < 10000)) {
      Self->Scale = *Scale;
      if (Self->DrawableID) ActionMsg(AC_Draw, Self->DrawableID, NULL);
      return(ERR_Okay);
   }
   else return(ERR_OutOfRange);
}

/*****************************************************************************

###FIELD###
Name:   Thickness
Short:  Defines the thickness of the Line in pixels.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The thickness of a line can be adjusted by writing a value to this field.
The default value is 1, which creates a thin line that is a single pixel wide.
Note that the thicker the line, the more time that will need to be taken to
draw it.</p>

###END###

*****************************************************************************/

static ERROR GET_Thickness(struct Line *Self, LONG *Value)
{
   if (Self->Pen) *Value = Self->Pen->TipSize;
   return(ERR_Okay);
}

static ERROR SET_Thickness(struct Line *Self, LONG *Value)
{
   if ((*Value < 1) OR (*Value > 20)) return(ERR_Failed);
   if (Self->Pen) Self->Pen->TipSize = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_X(struct Line *Self, struct Variable *Value)
{
   Self->Points[0].Flags &= ~LF_XFLAGS; /* Turn off existing x coordinate flags */

   if (Value->Type & FD_DOUBLE) Self->Points[0].XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->Points[0].XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Points[0].Flags |= LF_RELATIVE_X;

   if (Self->TotalPoints < 1) Self->TotalPoints = 1;
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_XEnd(struct Line *Self, struct Variable *Value)
{
   Self->Points[1].Flags &= ~LF_XFLAGS; /* Turn off existing x coordinate flags */

   if (Value->Type & FD_DOUBLE) Self->Points[1].XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->Points[1].XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Points[1].Flags |= LF_RELATIVE_X;

   if (Self->TotalPoints < 2) Self->TotalPoints = 2;
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_Y(struct Line *Self, struct Variable *Value)
{
   Self->Points[0].Flags &= ~LF_YFLAGS; /* Turn off existing y coordinate flags */

   if (Value->Type & FD_DOUBLE) Self->Points[0].YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->Points[0].YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Points[0].Flags |= LF_RELATIVE_Y;

   if (Self->TotalPoints < 1) Self->TotalPoints = 1;
   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_YEnd(struct Line *Self, struct Variable *Value)
{
   Self->Points[1].Flags &= ~LF_YFLAGS; /* Turn off existing y coordinate flags */

   if (Value->Type & FD_DOUBLE) Self->Points[1].YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->Points[1].YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Points[1].Flags |= LF_RELATIVE_Y;

   if (Self->TotalPoints < 2) Self->TotalPoints = 2;
   return(ERR_Okay);
}

/*****************************************************************************

                 "I see little demons, are they in my head?
                  I see little demons, fighting in my bed."

                       - King Diamond

*****************************************************************************/
